/* -*- mode: c++; c-basic-offset: 4; indent-tabs-mode: nil -*-

this file is part of rcssserver3D
Fri May 9 2003
Copyright (C) 2002,2003 Koblenz University
Copyright (C) 2003 RoboCup Soccer Server 3D Maintenance Group

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "kicknrun.h"

/*-------------------------------------------------------------------*/
/*!

*/
KickNRun::KickNRun()
    : Soccer()
{

}

/*-------------------------------------------------------------------*/
/*!

*/
KickNRun::~KickNRun()
{

}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickNRun::behave()
{
    TTeamIndex tidx = M_wm->getMyTeam();
    if ( tidx != TI_LEFT && tidx != TI_RIGHT )
    {
        GetLog()->Debug() << "Init failed:"
                          << M_wm->getTeamUnum()
                          << std::endl;
        doInitAgent();
        return;
    }
    TPlayMode my_kick_off = ( tidx == TI_LEFT
                              ? PM_KickOff_Left
                              : PM_KickOff_Right );
    TPlayMode my_kick_in  = ( tidx == TI_LEFT
                              ? PM_KickIn_Left
                              : PM_KickIn_Right );
    TPlayMode pm = M_wm->getPlayMode();
    switch ( pm ) {
    case PM_BeforeKickOff:
        playBeforeKickOff();
        break;
    case PM_KickOff_Left:
    case PM_KickOff_Right:
        if ( pm == my_kick_off )
        {
            playMyKickOff();
        }
        else
        {
            playTheirKickOff();
        }
        break;
    case PM_PlayOn :
        playPlayOn();
        break;
    case PM_KickIn_Left :
    case PM_KickIn_Right :
        if ( pm == my_kick_in )
        {
            playMyKickIn();
        }
        else
        {
            playTheirKickIn();
        }
        break;
    default:
        GetLog()->Debug() << "Unknown PM: " << pm << std::endl;
        playPlayOn();
        break;
    }

}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickNRun::playBeforeKickOff()
{
    WorldModel::FormationType ft = WorldModel::FT_initial;
    int unum = M_wm->getTeamUnum();
    doBeam( M_wm->M_formations[ft].player[unum].pos() );
    doDrive( salt::Vector3f( 0.0f, 0.0f, 0.0f ) );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickNRun::playMyKickOff()
{
    if ( M_wm->getTeamUnum() == 11 )
    {
        playPlayOn();
    }
    else
    {
        doStopMotor();
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickNRun::playTheirKickOff()
{
    // do nothing
    doStopMotor();
}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickNRun::playMyKickIn()
{
    // for now just kick and run
    playPlayOn( true );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickNRun::playTheirKickIn()
{
    // do nothing
    doStopMotor();
}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickNRun::playPlayOn( bool kickin )
{
    WorldModel::VisionObject my_vo_type = M_wm->getMyObject();
    salt::Vector3f my_pos = M_wm->getMyPosition();
    salt::Vector3f my_vel = M_wm->getObjectSpeed( my_vo_type );

    GetLog()->Debug() << "MYPOS= " << my_pos
                      << "\nMYVEL= " << my_vel
                      << std::endl;

    salt::Vector3f ball_pos = M_wm->getObjectPosition( WorldModel::VO_BALL );
    salt::Vector3f ball_vel = M_wm->getObjectSpeed( WorldModel::VO_BALL );

    GetLog()->Debug() << "BALLPOS = " <<  ball_pos
                      << "\nBALLVEL = " << ball_vel
                      << std::endl;
    ball_pos = ball_pos
        + ball_vel * M_wm->getNeedStepsSetTo( my_vo_type, WorldModel::VO_BALL );
    GetLog()->Debug() << "BALLFUTURE = " <<  ball_pos << std::endl;



    if ( M_wm->getTeamUnum() == 1 )
    {
        playPlayOnGoalie();
        return;
    }

    salt::Vector3f kick_pos = getKickPoint();
    salt::Vector3f dl2KickPos = kick_pos - M_wm->getMyPosition();

    if ( dl2KickPos.Length() < M_wm->getKickableDistance() )
    {
        if ( my_pos.x() < 0 )
        {
            doKick( 00.0f, 100.0f );
        }
        else if ( my_pos.x() < 25.0f )
        {
            doKick( 0.0f, 100.0f );
        }
        else
        {
            doKick( 0.0f, 100.0f );
        }
        return;
    }

    if ( M_wm->getFastestInSetTo( WorldModel::SET_TEAMMATE_NOGK,
                                  WorldModel::VO_BALL ) == my_vo_type )
    {
        doIntercept( kick_pos );
        return;
    }

    double dist_thr = ( my_pos - ball_pos ).Length() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    salt::Vector3f home_rel = ( M_wm->getStrategicPosition() - my_pos );
    double dist_diff = home_rel.Length();
    if ( dist_diff > dist_thr )
    {
        double rate = 60.0;
        if ( dist_diff < 10.0 ) rate = 50.0;
        else if ( dist_diff < 5.0 ) rate = 30.0;
        else if ( dist_diff < 3.0 ) rate = 10.0;

        doDrive( home_rel.Normalized() * rate );
        return;
    }

    doStopMotor();
}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickNRun::playPlayOnGoalie()
{
    salt::Vector3f my_pos = M_wm->getMyPosition();
    WorldModel::VisionObject my_vo_type = M_wm->getMyObject();
    salt::Vector3f my_vel = M_wm->getObjectSpeed( my_vo_type );
    salt::Vector3f ball_pos = M_wm->getObjectPosition( WorldModel::VO_BALL );

    salt::Vector3f kick_pos = getKickPoint();
    salt::Vector3f dl2KickPos = kick_pos - my_pos;

    if ( dl2KickPos.Length() < M_wm->getKickableDistance() )
    {
        doKick( 0.0f, 100.0f );
        return;
    }

    if ( M_wm->getFastestInSetTo( WorldModel::SET_TEAMMATE, WorldModel::VO_BALL ) == my_vo_type )
    {
        if ( ball_pos.x() < -30 && std::abs( ball_pos.y() ) < 10 )
        {
            if ( dl2KickPos.Length() > 0.5 )
            {
                doDrive( (kick_pos - my_pos).Normalized() * 100.0 );
            }
            else if ( dl2KickPos.Length() > 0.35 && my_vel.Length() > 0.2 )
            {
                doDrive( my_vel.Normalized() * -80.0 ); // break.
            }
            else if ( dl2KickPos.Length() > M_wm->getKickableDistance() )
            {
                doDrive( (kick_pos - my_pos).Normalized() * 20.0 );
            }
            else
            {
                doDrive( (kick_pos - my_pos).Normalized() * 20.0 );
            }
            return;
        }
    }
    /*
    const double miny = -M_wm->getGoalWidth() * 0.5 - 2.0;
    const double maxy = -miny;
    const double move_x = - M_wm->getFieldLength() * 0.5 + 3.0;
    double slope = ( ball_pos.x() - move_x ) / ( ball_pos.x() - move_x - 5.0 );
    double move_y = ball_pos.y() * slope;

    if ( move_y < miny ) move_y = miny;
    if ( move_y > maxy ) move_y = maxy;
    */
    salt::Vector3f home_pos( -52.0, 0.0, 0 );

    double dist_thr = ( my_pos - ball_pos ).Length() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    if ( ( home_pos - my_pos ).Length() > dist_thr )
    {
        doDrive( ( home_pos - my_pos ).Normalized() * 30.0 );
        return;
    }

    doStopMotor();
}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickNRun::doStopMotor()
{
    doDrive( salt::Vector3f( 0.0f, 0.0f, 0.0f ) );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickNRun::doGoToPoint( const salt::Vector3f& target_point,
                       const double& speed_rate )
{
    salt::Vector3f target_rel = target_point - M_wm->getMyPosition();
    double rate = std::min( speed_rate, 100.0 );
    rate = std::max( 0.0, rate );
    doDrive( target_rel.Normalized() * rate );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickNRun::doIntercept( const salt::Vector3f& kicking_pos )
{
    salt::Vector3f target_rel = kicking_pos - M_wm->getMyPosition();
    float len = target_rel.Length();

    GetLog()->Debug() << "Intercept: target_rel = " << target_rel
                      << "  target_dist = " << len
                      << "  MyKickable = " << M_wm->getKickableDistance()
                      << std::endl;

    if ( len > 0.5 )
    {
        // max power
        doDrive( target_rel.Normalized() * 100.0f );
        return;
    }

    salt::Vector3f my_vel = M_wm->getObjectSpeed( M_wm->getMyObject() );

    if ( len > 0.35 && my_vel.Length() > 0.2 )
    {
        GetLog()->Debug() << "Intercept: Try to stop "
                          << std::endl;
        doDrive( my_vel.Normalized() * -50.0 );
        return;
    }

    if ( len > M_wm->getKickableDistance() )
    {
        doDrive( target_rel.Normalized() * 20.0 );
        return;
    }

    doDrive( target_rel.Normalized() * 20.0 );
}

/*-------------------------------------------------------------------*/
/*!

*/
salt::Vector3f
KickNRun::getKickPoint()
{
    static double s_my_max_speed = 0.0;

    salt::Vector3f my_pos = M_wm->getMyPosition();
    salt::Vector3f my_vel = M_wm->getObjectSpeed( M_wm->getMyObject() );
    salt::Vector3f ball_pos = M_wm->getObjectPosition( WorldModel::VO_BALL );
    salt::Vector3f ball_vel = M_wm->getObjectSpeed( WorldModel::VO_BALL );

    salt::Vector3f shot_target( M_wm->getFieldLength() / 2.0, 0.0, 0.0 );

    if ( s_my_max_speed < my_vel.Length() )
    {
        s_my_max_speed = my_vel.Length();
        GetLog()->Debug() << "Updated my max speed = "
                          << s_my_max_speed
                          << std::endl;
    }

    // my max speed is about 0.32

    // ball decay sample
    // 0.902431434
    // 0.902371014
    // 0.903324807
    int steps = 0;
    salt::Vector3f tmp_ball_pos = ball_pos;
    salt::Vector3f tmp_ball_vel = ball_vel;
    double my_dash = 0.0;
    while ( steps < 100 )
    {
        my_dash += 0.3;
        tmp_ball_pos += tmp_ball_vel;
        tmp_ball_vel *= 0.9;
        ++steps;
        if ( ( tmp_ball_pos - my_pos ).Length() < my_dash + 1.0 )
        {
            break;
        }
    }

    ball_pos = ball_pos
        + ball_vel * ( 1 - std::pow( 0.9, steps ) ) / ( 1.0 - 0.9 );


    salt::Vector3f ball2goal = shot_target - ball_pos;

    if ( ball2goal.Length() < 15.0 )
    {
        // goale is near
        // return shoot point
        return ball_pos - ( ball2goal.Normalized() * M_wm->getKickableDistance() );
    }
#if 1
    // search pass receiver
    for ( int i = 1; i <= 11; ++i )
    {
        if ( i == M_wm->getTeamUnum() ) continue;

        WorldModel::VisionObject vo
            = static_cast< WorldModel::VisionObject >
            ( static_cast< int >( WorldModel::VO_OUR0 ) + i );

        salt::Vector3f pass_pos = M_wm->getObjectPosition( vo );
        if ( pass_pos.x() - my_pos.x() > 5.0 )
        {
            double len = ( pass_pos - my_pos ).Length();
            if ( 10.0 < len && len < 20.0 )
            {
                pass_pos += salt::Vector3f( 5.0, 0.0, 0.0 );
                return ( ball_pos
                         - ( ( pass_pos - ball_pos ).Normalized()
                             * M_wm->getKickableDistance() ) );
            }
        }
    }
#endif

    return ( ball_pos
             - ( ball2goal.Normalized() * M_wm->getKickableDistance() ) );
}
