/* -*- mode: c++; c-basic-offset: 4; indent-tabs-mode: nil -*-

this file is part of rcssserver3D
Fri May 9 2003
Copyright (C) 2002,2003 Koblenz University
Copyright (C) 2003 RoboCup Soccer Server 3D Maintenance Group

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "worldmodel.h"

/*-------------------------------------------------------------------*/
/*!

*/
WorldModel::WorldModel()
{
    M_field_length = 105.0;
    M_field_width = 68.0;
    M_field_height = 20.0;
    M_goal_width = 7.32;
    M_goal_depth = 2.0;
    M_goal_height = 2.44;
    M_border_size = 4.0f;

    // agent parameter
    M_agent_mass = 75;
    M_agent_radius = 0.22;
    M_agent_max_speed = 10;

    // ball parameter
    M_ball_radius = 0.111;
    M_ball_mass = 0.4;

    // team information
    M_team_index = TI_NONE;
    M_team_unum = 0;


    M_my_pos_v_enable = false;

    Formation* fp = &( M_formations[FT_initial] );
    int i = 1;
    fp->player[i++].setValues( -50.0,	  0.0, 0.0, 0.0, -49.0,	0.0 );
    fp->player[i++].setValues( -17.0,	 -4.0, 0.0, 0.0, -45.0,	0.0 );
    fp->player[i++].setValues( -17.0,	  4.0, 0.0, 0.0, -45.0,	0.0 );
    fp->player[i++].setValues( -16.0,	-10.0, 0.0, 0.0, -45.0,	0.0 );
    fp->player[i++].setValues( -16.0,  10.0, 0.0, 0.0, -45.0,	0.0 );
    fp->player[i++].setValues(  -9.0,   0.0, 0.0, 0.0, -45.0,	0.0 );
    fp->player[i++].setValues(  -5.0, -10.0, 0.0, 0.0, -45.0,	0.0 );
    fp->player[i++].setValues(  -5.0,  10.0, 0.0, 0.0, -45.0,	0.0 );
    fp->player[i++].setValues(  -1.0, -22.0, 0.0, 0.0, -45.0,	0.0 );
    fp->player[i++].setValues(  -1.0,  22.0, 0.0, 0.0, -45.0,	0.0 );
    fp->player[i++].setValues(  -1.0,   0.0, 0.0, 0.0, -45.0,	0.0 );

    fp = &( M_formations[FT_433_offensive] );
    i = 1;
    //                             x      y attx  atty    minx   maxx   miny   maxy  behindball
    fp->player[i++].setValues( -50.0,   0.0, 0.0,  0.0,  -52.0, -48.0  -10.0,  10.0, true ); //GK
    fp->player[i++].setValues( -16.5,  -4.0, 0.5,  0.3,  -44.5,   0.0, -18.0,   8.0, true ); //L-CB
    fp->player[i++].setValues( -16.5,   4.0, 0.5,  0.3,  -44.5,   0.0,  -8.0,  18.0, true ); //R-CB
    fp->player[i++].setValues( -14.5, -10.0, 0.5,  0.3,  -47.0,   2.0, -32.0,  -5.0, true ); //L-SB
    fp->player[i++].setValues( -14.5,  10.0, 0.5,  0.3,  -47.0,   2.0,   5.0,  32.0, true ); //R-SB
    fp->player[i++].setValues(  -5.0,   0.0, 0.5,  0.35, -36.0,  18.0, -18.0,  18.0 ); // DMF
    fp->player[i++].setValues(   0.0, -11.0, 0.5,  0.4,  -31.0,  35.0, -20.0,  -1.0 ); // L-OMF
    fp->player[i++].setValues(   0.0,  11.0, 0.5,  0.4,  -31.0,  35.0,   1.0,  20.0 ); // R-OMF
    fp->player[i++].setValues(  15.0, -18.0, 0.6,  0.5,   -9.0,  44.0, -25.0,  -7.0 );  // L-FW
    fp->player[i++].setValues(  15.0,  18.0, 0.6,  0.5,   -9.0,  44.0,   7.0,  25.0 );  // L-FW
    fp->player[i++].setValues(  18.0,   0.0, 0.6,  0.5,  -12.0,  44.0, -12.0,  12.0 );  // CFW
}

/*-------------------------------------------------------------------*/
/*!

*/
WorldModel::~WorldModel()
{

}

/*-------------------------------------------------------------------*/
/*!

*/
void
WorldModel::setupGameStateMap()
{
    M_playmode_map.clear();

    M_playmode_map[STR_PM_BeforeKickOff] = PM_BeforeKickOff;
    M_playmode_map[STR_PM_KickOff_Left]  = PM_KickOff_Left;
    M_playmode_map[STR_PM_KickOff_Right] = PM_KickOff_Right;
    M_playmode_map[STR_PM_PlayOn]        = PM_PlayOn;
    M_playmode_map[STR_PM_KickIn_Left]   = PM_KickIn_Left;
    M_playmode_map[STR_PM_KickIn_Right]  = PM_KickIn_Right;
    M_playmode_map[STR_PM_Goal_Left]     = PM_Goal_Left;
    M_playmode_map[STR_PM_Goal_Right]    = PM_Goal_Right;
    M_playmode_map[STR_PM_GameOver]      = PM_GameOver;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
WorldModel::setupVisionObjectMap()
{
    M_vision_object_map.clear();

    // <ObjectName>+<Id> -> VisionObject
    M_vision_object_map["Flag1_l"] = VO_FLAG1L;
    M_vision_object_map["Flag1_r"] = VO_FLAG1R;
    M_vision_object_map["Flag2_l"] = VO_FLAG2L;
    M_vision_object_map["Flag2_r"] = VO_FLAG2R;
    M_vision_object_map["Goal1_l"] = VO_GOAL1L;
    M_vision_object_map["Goal1_r"] = VO_GOAL1R;
    M_vision_object_map["Goal2_l"] = VO_GOAL2L;
    M_vision_object_map["Goal2_r"] = VO_GOAL2R;
    M_vision_object_map["Ball"]    = VO_BALL;

    M_vision_object_map["Our1"]    = VO_OUR1;
    M_vision_object_map["Our2"]    = VO_OUR2;
    M_vision_object_map["Our3"]    = VO_OUR3;
    M_vision_object_map["Our4"]    = VO_OUR4;
    M_vision_object_map["Our5"]    = VO_OUR5;
    M_vision_object_map["Our6"]    = VO_OUR6;
    M_vision_object_map["Our7"]    = VO_OUR7;
    M_vision_object_map["Our8"]    = VO_OUR8;
    M_vision_object_map["Our9"]    = VO_OUR9;
    M_vision_object_map["Our10"]    = VO_OUR10;
    M_vision_object_map["Our11"]    = VO_OUR11;

    M_vision_object_map["Opp1"]    = VO_OPP1;
    M_vision_object_map["Opp2"]    = VO_OPP2;
    M_vision_object_map["Opp3"]    = VO_OPP3;
    M_vision_object_map["Opp4"]    = VO_OPP4;
    M_vision_object_map["Opp5"]    = VO_OPP5;
    M_vision_object_map["Opp6"]    = VO_OPP6;
    M_vision_object_map["Opp7"]    = VO_OPP7;
    M_vision_object_map["Opp8"]    = VO_OPP8;
    M_vision_object_map["Opp9"]    = VO_OPP9;
    M_vision_object_map["Opp10"]    = VO_OPP10;
    M_vision_object_map["Opp11"]    = VO_OPP11;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
WorldModel::ConstructInternal()
{
    M_parser_ptr
        = boost::shared_dynamic_cast< oxygen::BaseParser >
        ( GetCore()->New( "SexpParser" ) );

    if ( M_parser_ptr.get() == 0 )
    {
        GetLog()->Error()
            << "ERROR: (WorldModel) create sexpparser parser instance\n.";

        return false;
    }

    setupGameStateMap();
    setupVisionObjectMap();

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
WorldModel::parseTeamIndex( const oxygen::Predicate& predicate )
{
    std::string team;
    if ( ! predicate.GetValue( predicate.begin(), "team", team ) )
    {
        return;
    }

    // parse team index
    if ( team == "left" )
    {
        M_team_index = TI_LEFT;
    }
    else if ( team == "right" )
    {
        M_team_index = TI_RIGHT;
    }
    else
    {
        M_team_index = TI_NONE;
        GetLog()->Debug()
            << "(WorldModel) received TeamIndex TI_NONE\n";
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
WorldModel::parsePlayMode( const oxygen::Predicate& predicate )
{
    std::string mode;
    if ( ! predicate.GetValue( predicate.begin(), "playmode", mode ) )
    {
        return;
    }

    // parse play mode
    TPlayModeMap::iterator iter = M_playmode_map.find( mode );
    if ( iter != M_playmode_map.end() )
    {
        M_playmode = (TPlayMode)(*iter).second;
    }
    else
    {
        GetLog()->Debug()
            << "ERROR: (WorldModel) could not parse playmode "
            << mode << "\n";
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
WorldModel::parseGameState( const oxygen::Predicate& predicate )
{
    parseTeamIndex( predicate );
    parsePlayMode( predicate );

    predicate.GetValue( predicate.begin(), "FieldLength",   M_field_length );
    predicate.GetValue( predicate.begin(), "FieldWidth",    M_field_width );
    predicate.GetValue( predicate.begin(), "FieldHeigth",   M_field_height );
    predicate.GetValue( predicate.begin(), "GoalWidth",     M_goal_width );
    predicate.GetValue( predicate.begin(), "GoalDepth",     M_goal_depth );
    predicate.GetValue( predicate.begin(), "GoalHeight",    M_goal_height );
    predicate.GetValue( predicate.begin(), "BorderSize",    M_border_size );
    predicate.GetValue( predicate.begin(), "time",          M_time );
    predicate.GetValue( predicate.begin(), "unum",          M_team_unum );
    predicate.GetValue( predicate.begin(), "AgentRadius",   M_agent_radius );
    predicate.GetValue( predicate.begin(), "AgentMass",     M_agent_mass );
    predicate.GetValue( predicate.begin(), "AgentMaxSpeed", M_agent_max_speed );
    predicate.GetValue( predicate.begin(), "BallRadius",    M_ball_radius );
    predicate.GetValue( predicate.begin(), "BallMass",      M_ball_mass );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
WorldModel::parseObjectVision( const oxygen::Predicate& predicate )
{
    for ( oxygen::Predicate::Iterator iter(predicate);
          iter != iter.end();
          ++iter )
    {
        // extract the element as a parameter list
        oxygen::Predicate::Iterator paramIter = iter;
        if ( ! predicate.DescentList( paramIter ) )
        {
            continue;
        }

        // read the object name
        std::string name;
        if ( ! predicate.GetValue( paramIter, name ) )
        {
            continue;
        }

        if ( name == "Player" )
        {
            std::string strTeam;
            if ( predicate.GetValue( paramIter, "team", strTeam ) )
            {
                if ( strTeam == getTeamName() )
                {
                    name = "Our";
                }
                else
                {
                    name = "Opp";
                }
            }
        }

        // try read the 'id' section
        std::string strId;
        if ( predicate.GetValue( paramIter,"id", strId ) )
        {
            name += strId;
        }

        // try to lookup the VisionObject
        TVisionObjectMap::iterator iter = M_vision_object_map.find( name );
        if ( iter == M_vision_object_map.end() )
        {
            GetLog()->Debug() << "cant find in M_vision_object_map: " << name
                              << std::endl;
            continue;
        }

        VisionObject vo = (*iter).second;

        // find  to the 'pol' entry in the object's section
        oxygen::Predicate::Iterator polIter = paramIter;
        if ( ! predicate.FindParameter( polIter, "pol" ) )
        {
            continue;
        }

        // read the position vector
        VisionSense sense;
        if ( ! predicate.AdvanceValue( polIter, sense.distance_ )
             || ! predicate.AdvanceValue( polIter, sense.theta_ )
             || ! predicate.AdvanceValue( polIter, sense.phi_ )
             )
        {
            continue;
        }

        // update the vision map
        M_vision_map[vo] = sense;
    }

    VisionSense my_pol;
    my_pol.distance_ = 0;
    my_pol.theta_ = 0;
    my_pol.phi_ = 0;
    M_vision_map[static_cast< VisionObject >( static_cast< int >( VO_OUR0 )
                                              + getTeamUnum())]
        = my_pol;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
WorldModel::parseVision( const oxygen::Predicate& predicate )
{
    parseObjectVision( predicate );

    //
    // get our position. The current implementation of the worldmodel
    // relies on the VisionPerceptor that deliverd our position. In
    // future releases this should be replaced by a self localization
    // algorithm using the relative positions of some flags.
    //

    // find the PerfectVision data about the object
    oxygen::Predicate::Iterator iter( predicate );

    // advance to the section about object 'name'
    if ( ! predicate.FindParameter( iter, "mypos" ) )
    {
        M_my_pos_v_enable = false;
        return;
    }

    // read my position
    VisionSense sense;

    predicate.GetValue( iter, M_my_pos_v );
    M_my_pos_v_enable = true;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
WorldModel::parse( const std::string& message )
{
    boost::shared_ptr< oxygen::PredicateList >
        pred_list = M_parser_ptr->Parse( message );

    if ( pred_list.get() == 0 )
    {
        return;
    }

    oxygen::PredicateList& list = *pred_list;

    M_position_map.clear();
    M_position_map[VO_BALL] = getObjectPosition( VO_BALL );
    for ( VisionObject vo = VO_OUR1;
          vo <= VO_OUR11;
          vo = static_cast< VisionObject >( static_cast< int >( vo ) + 1 ) )
    {
        M_position_map[vo] = getObjectPosition( vo );
    }

    for ( VisionObject vo = VO_OPP1;
          vo <= VO_OPP11;
         vo = static_cast< VisionObject >( static_cast< int >( vo ) + 1 ) )
    {
        M_position_map[vo] = getObjectPosition( vo );
    }

    for ( oxygen::PredicateList::TList::const_iterator iter = list.begin();
          iter != list.end();
          ++iter )
    {
        const oxygen::Predicate& predicate = (*iter);

        if ( predicate.name == "GameState" )
        {
            parseGameState( predicate );
            continue;
        }

        // check for the Vision perceptor
        if ( predicate.name == "Vision" )
        {
            parseVision( predicate );
            continue;
        }

        GetLog()->Debug()
            << "(WorldModel) skipping unknown predicate "
            << predicate.name << "\n";
    }

    // update our position
    calcPlayerPosition();
}

/*-------------------------------------------------------------------*/
/*!

 */
WorldModel::VisionSense
WorldModel::getVisionSense( VisionObject obj )
{
    TVisionMap::iterator iter = M_vision_map.find( obj );

    if ( iter == M_vision_map.end() )
    {
        GetLog()->Error()
            << "ERROR: (WorldMode::getVisionSense) unknown VisionObject "
            << obj << "\n";

        return VisionSense();
    }

    return (*iter).second;
}

/*-------------------------------------------------------------------*/
/*!

 */

salt::Vector3f
WorldModel::getPosition( VisionSense sense )
{
    return M_my_pos + getDriveVec( sense ) * sense.distance_;
}

/*-------------------------------------------------------------------*/
/*!

 */
salt::Vector3f
WorldModel::getObjectPosition( VisionObject obj )
{
    return getPosition( getVisionSense( obj ) );
}

/*-------------------------------------------------------------------*/
/*!

 */
salt::Vector3f
WorldModel::getObjectSpeed( VisionObject obj )
{
    if ( getPlayMode() != PM_PlayOn )
    {
        return salt::Vector3f( 0, 0, 0 );
    }
    return getPosition( getVisionSense( obj ) ) - M_position_map[obj];
}

/*-------------------------------------------------------------------*/
/*!

 */
WorldModel::VisionObject
WorldModel::getMyObject()
{
    return static_cast< VisionObject >( static_cast< int >( VO_OUR0 )
                                        + getTeamUnum() );
}

/*-------------------------------------------------------------------*/
/*!

 */
salt::Vector3f
WorldModel::getDriveVec( const WorldModel::VisionSense& vision )
{
#if 0
    return salt::Vector3f
        (
         vision.distance_ * salt::gCos( salt::gDegToRad( vision.theta_ ) )
         * salt::gSin( salt::gDegToRad( 90.0 - vision.phi_ ) ),

         vision.distance_ * salt::gSin( salt::gDegToRad( vision.theta_ ) )
         * salt::gSin( salt::gDegToRad( 90.0 - vision.phi_ ) ),

         vision.distance_ * salt::gCos( salt::gDegToRad( 90.0 - vision.phi_ ) )
         );
#else
    // TsubameGaeshi version
    return salt::Vector3f
        (
         salt::gCos( salt::gDegToRad( vision.theta_ ) )
         * salt::gSin( salt::gDegToRad( 90.0 - vision.phi_ ) ),

         salt::gSin( salt::gDegToRad( vision.theta_ ) )
         * salt::gSin( salt::gDegToRad( 90.0 - vision.phi_ ) ),

         salt::gCos( salt::gDegToRad( 90.0 - vision.phi_ ) )
         );
#endif
}

/*-------------------------------------------------------------------*/
/*!

 */
void
WorldModel::calcPlayerPosition()
{
    WorldModel::VisionSense fs;
    salt::Vector3f flag_pos;
    salt::Vector3f my_pos;

    int sign = ( getMyTeam() == TI_RIGHT
                 ? -1
                 : 1 );
    //F1L
    fs = getVisionSense( WorldModel::VO_FLAG1L );
    flag_pos = salt::Vector3f( -getFieldLength() * 0.5f,
                               -getFieldWidth() * 0.5f,
                               0.0f );
    my_pos = flag_pos - getDriveVec( fs ) * sign * fs.distance_;

    //F2L
    fs = getVisionSense( WorldModel::VO_FLAG2L );
    flag_pos = salt::Vector3f( -getFieldLength() * 0.5f,
                               getFieldWidth() * 0.5f,
                               0.0f );
    my_pos += flag_pos - getDriveVec( fs ) * sign * fs.distance_;

    //F1R
    fs = getVisionSense( WorldModel::VO_FLAG1R );
    flag_pos = salt::Vector3f( getFieldLength() * 0.5f,
                               -getFieldWidth() * 0.5f,
                               0.0f );
    my_pos += flag_pos - getDriveVec( fs ) * sign * fs.distance_;

    //F2R
    fs = getVisionSense( WorldModel::VO_FLAG2R );
    flag_pos = salt::Vector3f( getFieldLength() * 0.5f,
                         getFieldWidth() * 0.5f,
                         0.0f );
    my_pos += flag_pos - getDriveVec(fs) * sign * fs.distance_;

    M_my_pos = my_pos / 4;

    GetLog()->Debug() << "MYPOS  :"
                      << M_my_pos[0] << " "
                      << M_my_pos[1] << " "
                      << M_my_pos[2] << "\n";
    // if right side, reverse
    M_my_pos  = M_my_pos * sign;
    M_my_pos_v = M_my_pos_v * sign;

    GetLog()->Debug() << "MYPOS  :"
                      << M_my_pos[0] << " "
                      << M_my_pos[1] << " "
                      << M_my_pos[2] << "\n";
    if ( M_my_pos_v_enable )
    {
        GetLog()->Debug() << "MYPOSV :"
                          << M_my_pos_v[0] << " "
                          << M_my_pos_v[1] << " "
                          << M_my_pos_v[2] << "\n";
        GetLog()->Debug() << "MYPOS error :"
                          << ( M_my_pos_v - M_my_pos ).Length() << "\n";
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
int
WorldModel::getNeedStepsSetTo( VisionObject po, VisionObject vo )
{
    salt::Vector3f player_pos = getObjectPosition( po );
    salt::Vector3f object_pos = getObjectPosition( vo );
    salt::Vector3f object_vel = getObjectSpeed( vo );

#if 1
    int step = 0;
    while ( step < 20
            && ( object_pos - player_pos ).Length() > 0.3 * step )
    {
        object_pos += object_vel;
        ++step;
    }
    return step;
#else
    int s = 0;
    for ( s = 0;
          ( object_pos - player_pos ).Length() > 0.3 * s && s < 20;
          ++s )
    {
        object_pos += object_vel;
    }
    return s;
#endif
}

#if 1
/*-------------------------------------------------------------------*/
/*!

 */
WorldModel::VisionObject
WorldModel::getFastestInSetTo( PlayersSet ps, VisionObject vo )
{
    VisionObject result = VO_OUR0;
    VisionObject obj;
    salt::Vector3f obj_pos = getObjectPosition( vo );

    double dist;
    double dist2;

    if ( ps == SET_TEAMMATE )
    {
        result = VO_OUR1;

        dist = getNeedStepsSetTo( result, vo ) * 100
            + ( obj_pos - getObjectPosition( result ) ).Length();

        for ( int i = static_cast< int >( VO_OUR2 );
              i <= static_cast< int >( VO_OUR11 );
              ++i )
        {
            obj = static_cast< VisionObject >( i );
            dist2 = getNeedStepsSetTo( obj, vo ) * 100
                + ( obj_pos - getObjectPosition( obj ) ).Length();
            if ( dist > dist2 )
            {
                dist = dist2;
                result = obj;
            }
        }
    }

    if ( ps == SET_TEAMMATE_NOGK )
    {
        result = VO_OUR1;

        dist = getNeedStepsSetTo( result, vo ) * 100
            + ( obj_pos - getObjectPosition( result ) ).Length();

        for ( int i = static_cast< int >( VO_OUR2 );
              i <= static_cast< int >( VO_OUR11 );
              ++i )
        {
            obj = static_cast< VisionObject >( i );
            dist2 = getNeedStepsSetTo( obj, vo ) * 100
                + ( obj_pos - getObjectPosition( obj ) ).Length();
            if ( dist > dist2 )
            {
                dist = dist2;
                result = obj;
            }
        }
    }

    if ( ps == SET_OPPONENT )
    {
        result = VO_OPP1;
        dist = getNeedStepsSetTo( result, vo ) * 100
            + ( obj_pos - getObjectPosition( result ) ).Length();

        for ( int i = static_cast< int >( VO_OPP2 );
              i <= static_cast< int >( VO_OPP2 );
              ++i )
        {
            obj = static_cast< VisionObject >( i );
            dist2 = getNeedStepsSetTo( obj, vo ) * 100
                + ( obj_pos - getObjectPosition( obj ) ).Length();
            if ( dist > dist2 )
            {
                dist = dist2;
                result = obj;
            }
        }
    }

    if ( ps == SET_ALL )
    {
        result = VO_OUR1;
        dist = getNeedStepsSetTo( result, vo ) * 100
            + ( obj_pos - getObjectPosition( result ) ).Length();

        for ( int i = static_cast< int >( VO_OUR2 );
              i <= static_cast< int >( VO_OUR11 );
              ++i )
        {
            obj = static_cast< VisionObject >( i );
            dist2 = getNeedStepsSetTo( obj, vo ) * 100
                + ( obj_pos - getObjectPosition( obj ) ).Length();
            if ( dist > dist2 )
            {
                dist = dist2;
                result = obj;
            }
        }

        for ( int i = static_cast< int >( VO_OPP1 );
              i <= static_cast< int >( VO_OPP11 );
              ++i )
        {
            obj = static_cast< VisionObject >( i );
            dist2 = getNeedStepsSetTo( obj, vo ) * 100
                + ( obj_pos - getObjectPosition( obj ) ).Length();
            if ( dist > dist2 )
            {
                dist = dist2;
                result = obj;
            }
        }
    }

    return result;
}

#else
WorldModel::VisionObject
WorldModel::getFastestInSetTo( PlayersSet ps, VisionObject vo )
{
    salt::Vector3f obj_pos = getObjectPosition( vo );
    VisionObject result;
    double dist;
    double dist2;

    if ( ps == SET_TEAMMATE )
    {
        result = VO_OUR1;
        dist = ( obj_pos - getObjectPosition( result ) ).Length();

        for ( int i = static_cast< int >( VO_OUR2 );
              i <= static_cast< int >( VO_OUR11 );
              ++i )
        {
            dist2 = ( obj_pos - getObjectPosition( static_cast< VisionObject >( i ) ) ).Length();
            if ( dist > dist2 )
            {
                dist = dist2;
                result = static_cast< VisionObject >( i );
            }
        }
    }

    if ( ps == SET_TEAMMATE_NOGK )
    {
        result = VO_OUR2;
        dist = ( obj_pos - getObjectPosition( result ) ).Length();

        for ( int i = static_cast< int >( VO_OUR3 );
              i <= static_cast< int >( VO_OUR11 );
              ++i )
        {
            dist2 = ( obj_pos - getObjectPosition( static_cast< VisionObject >( i ) ) ).Length();
            if ( dist > dist2 )
            {
                dist = dist2;
                result = static_cast< VisionObject >( i );
            }
        }
    }

    if ( ps == SET_OPPONENT )
    {
        result = VO_OPP1;
        dist = ( obj_pos - getObjectPosition( result ) ).Length();

        for ( int i = static_cast< int >( VO_OPP2 );
              i <= static_cast< int >( VO_OPP2 );
              ++i )
        {
            dist2 = ( obj_pos - getObjectPosition( static_cast< VisionObject >( i ) ) ).Length();
            if ( dist > dist2 )
            {
                dist = dist2;
                result = static_cast< VisionObject >( i );
            }
        }
    }

    if ( ps == SET_ALL )
    {
        result = VO_OUR1;
        dist = ( obj_pos - getObjectPosition( result ) ).Length();

        for ( int i = static_cast< int >( VO_OUR2 );
              i <= static_cast< int >( VO_OUR11 );
              ++i )
        {
            dist2 = ( obj_pos - getObjectPosition( static_cast< VisionObject >( i ) ) ).Length();
            if ( dist > dist2 )
            {
                dist = dist2;
                result = static_cast< VisionObject >( i );
            }
        }

        for ( int i = static_cast< int >( VO_OPP1 );
              i <= static_cast< int >( VO_OPP11 );
              ++i )
        {
            dist2 = ( obj_pos - getObjectPosition( static_cast< VisionObject >( i ) ) ).Length();
            if ( dist > dist2 )
            {
                dist = dist2;
                result = static_cast< VisionObject >( i );
            }
        }
    }

    return result;
}
#endif

/*-------------------------------------------------------------------*/
/*!

 */
salt::Vector3f
WorldModel::getStrategicPosition()
{
    double x, y;
    // get the home position and calculate the associated strategic position
    const FormationParam& param
        = M_formations[FT_433_offensive].player[getTeamUnum()];
    salt::Vector3f ball_pos = getObjectPosition( VO_BALL );

    x = param.pos().x() + ball_pos.x() * param.getAttract().x();
    y = param.pos().y() + ball_pos.y() * param.getAttract().y();

    // when behind ball is set, do not move to point in front of ball
    if ( param.isBehindBall()
         && x > ball_pos.x() )
    {
        x = ball_pos.x();
    }

    if ( x > param.getMaxX() ) x = param.getMaxX();
    if ( x < param.getMinX() ) x = param.getMinX();

    if ( y > param.getMaxY() ) y = param.getMaxY();
    if ( y < param.getMinY() ) x = param.getMinY();

    return salt::Vector3f( x, y, 0 );
}

/*-------------------------------------------------------------------*/
/*!

 */
salt::Vector3f
WorldModel::getFromMyPosTo( salt::Vector3f pos )
{
    return ( M_team_index == TI_LEFT
             ? pos - M_my_pos
             : pos + M_my_pos );
}
